open Algebra

(******************************************************************************)
(* fractions ******************************************************************)
(******************************************************************************)

module type Fraction = sig
  module Numerator : Multipliable
  include Dividable

  val from_numerator : Numerator.t -> t
  val to_pair        : t -> Numerator.t * Numerator.t
end

module type OrdFraction = sig
  include Fraction
  include Ordered with type t := t
end

module Fraction    (M : Multipliable) : Fraction    with type Numerator.t = M.t
module OrdFraction (O : Ordered)      : OrdFraction with type Numerator.t = O.t

(******************************************************************************)
(* vectors   ******************************************************************)
(******************************************************************************)

module type Vector2 = sig
  module Scalar : Dividable
  include Addable

  val dot   : t -> t -> Scalar.t
  val cross : t -> t
  val norm2 : t -> Scalar.t
  val smult : Scalar.t -> t -> t
  val sdiv  : t -> Scalar.t -> t
end

module Vector2 (S : Number) : Vector2 with type Scalar.t = S.t

(******************************************************************************)
(* payoff    ******************************************************************)
(******************************************************************************)

(* we define all of the following modules, in just a few lines of code.  The
   type signatures of these modules combined contain ~150 lines of code (derived
   by cleaning up the output of #use "impls.ml").  The implementation contains
   ~120 lines of code (derived by stripping comments and blank lines from
   impls.ml). 150 functions in 120 lines, not bad! Even better when future
   implementations and extensions can reuse most of the code. *)

module Int          : Ordered with type t = int
module BigInt       : Ordered
module Float        : Number  with type t = float
module Rational     : Dividable
module BigRational  : Dividable
module Vec2OfFloat  : Vector2 with type Scalar.t = float
module Vec2OfRat    : Vector2 with type Scalar.t = Rational.t
module Vec2OfBigRat : Vector2 with type Scalar.t = BigRational.t


